package theorems

trait Formula {
  def symbol: Char
//  def sort: Formula
}

trait Atom extends Formula {
  def isPositive: Boolean
  override def toString = "" + symbol
  def sort: Atom = this
}

object PosAtom extends Atom { def isPositive = true;  def symbol = Formulas.positiveAtomSymbol }
object NegAtom extends Atom { def isPositive = false; def symbol = Formulas.negativeAtomSymbol }

object Atom {
  def apply(symbol: Char) = if (symbol == Formulas.positiveAtomSymbol) PosAtom else NegAtom
}

abstract class Connective(children: Vector[Formula], override val symbol: Char) extends Formula {
  def arity = children.size
  override def toString = symbol + children.map(_.toString).mkString("(",",",")")
}

object Connective {
  def apply(symbol: Char, children: Vector[Formula]) = symbol match {
    case Formulas.tensorSymbol => Tensor(children)
    case Formulas.parSymbol    =>    Par(children)
    case Formulas.withSymbol   =>   With(children)
    case Formulas.plusSymbol   =>   Plus(children)
  }
}

trait Additive extends Connective
trait Multiplicative extends Connective
trait Positive extends Connective
trait Negative extends Connective

case class Tensor(children: Vector[Formula]) extends Connective(children,Formulas.tensorSymbol) with Multiplicative with Positive
case class Par   (children: Vector[Formula]) extends Connective(children,Formulas.parSymbol   ) with Multiplicative with Negative
case class With  (children: Vector[Formula]) extends Connective(children,Formulas.withSymbol  ) with Additive       with Negative
case class Plus  (children: Vector[Formula]) extends Connective(children,Formulas.plusSymbol  ) with Additive       with Positive


object Formulas {

  val positiveAtomSymbol = '1'
  val negativeAtomSymbol = '0'
  val tensorSymbol = '@'
  val parSymbol = '#'
  val withSymbol = '&'
  val plusSymbol = '+'

  val isAtomSymbol = Set(positiveAtomSymbol,negativeAtomSymbol)
  val isConnectiveSymbol = Set(tensorSymbol,parSymbol,withSymbol,plusSymbol)

  case class ParsedFormulaFromHead(formula: Formula, remainder: String)

  case class ParsedChildrenFromHead(children: Vector[Formula], remainder: String)
  
  def parseFromFront(s: String): ParsedFormulaFromHead = {
    val symbol: Char = s(0)
    if (isAtomSymbol(symbol)) ParsedFormulaFromHead(Atom(symbol),s.tail)  // tail either starts with comma ',' or is empty
    else {
//      println("    isConnectiveSymbol "+symbol)
      val parsedChildren = parseChildrenInParensFromHead(s.tail) // tail starts "(", remainder either starts comma ',' or is empty
      ParsedFormulaFromHead(Connective(symbol,parsedChildren.children),parsedChildren.remainder)
    }
  }
  
  def parseChildrenInParensFromHead(s: String): ParsedChildrenFromHead = {
    var remainder = s
    var children = Vector[Formula]() 
    while (remainder.head != ')') {  // starts with comma or open '('
      val parsedChild = parseFromFront(remainder.tail)  // tail => drop leading '(' or comma
      remainder = parsedChild.remainder
      children :+= parsedChild.formula
    }
    ParsedChildrenFromHead(children,remainder.tail) // _.tail => drop the ')'   hence remainder.tail starts with comma or is empty
  }

  def parse(s: String): Formula = parseFromFront(s).formula

  def testParse: Unit = {
    val cases = List(
      (""+positiveAtomSymbol,PosAtom)
      ,
      (""+negativeAtomSymbol,NegAtom)
      ,
      (tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+")",Tensor(Vector(NegAtom,PosAtom)))
      ,
      (tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+","+negativeAtomSymbol+")",Tensor(Vector(NegAtom,PosAtom,NegAtom)))
      ,
      (withSymbol+"("+tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+","+negativeAtomSymbol+"),"+plusSymbol+"("+positiveAtomSymbol+","+negativeAtomSymbol+"))",
        With(Vector(Tensor(Vector(NegAtom,PosAtom,NegAtom)),Plus(Vector(PosAtom,NegAtom)))))
    )
    for ((input,expect) <- cases) {
      println()
      println("input  "+input)
      println("expect "+expect)
      val got = parse(input)
      println("got    "+got)
      require(expect == got)
    }
  }

  def main(args: Array[String]) {
    testParse
  }

}
